from collections.abc import Sequence
from typing import TypeVar

from pydantic_ai.agent import Agent, AgentRunResult
from pydantic_ai.mcp import MCPServer
from pydantic_ai.messages import UserContent
from pydantic_ai.tools import Tool

from lightblue_ai.mcps import get_mcp_servers
from lightblue_ai.models import infer_model
from lightblue_ai.prompts import get_system_prompt
from lightblue_ai.settings import Settings
from lightblue_ai.tools.manager import LightBlueToolManager

T = TypeVar("T")


class LightBlueAgent[T]:
    def __init__(
        self,
        model_name: str | None = None,
        system_prompt: str | None = None,
        result_type: T = str,
        result_tool_name: str = "final_result",
        result_tool_description: str | None = None,
        tools: list[Tool] | None = None,
        mcp_servers: list[MCPServer] | None = None,
    ):
        self.tool_manager = LightBlueToolManager()
        self.settings = Settings()
        tools = tools or []
        mcp_servers = mcp_servers or []

        self.agent = Agent(
            infer_model(model_name or self.settings.default_model),
            result_type=result_type,
            result_tool_name=result_tool_name,
            result_tool_description=result_tool_description,
            system_prompt=system_prompt or get_system_prompt(),
            tools=[*tools, *self.tool_manager.get_all_tools()],
            mcp_servers=[*mcp_servers, *get_mcp_servers()],
        )

    async def run(self, user_prompt: str | Sequence[UserContent]) -> AgentRunResult[T]:
        async with self.agent.run_mcp_servers():
            return await self.agent.run(user_prompt)



from typing import Annotated

from pydantic import Field
from pydantic_ai.tools import Tool

from lightblue_ai.tools.base import LightBlueTool, Scope
from lightblue_ai.tools.extensions import hookimpl


class ThinkingTool(LightBlueTool):
    """https://www.anthropic.com/engineering/claude-think-tool"""

    def __init__(self):
        self.scopes = [Scope.read]

    async def _think(
        self,
        thought: Annotated[str, Field(description="A thought to think about.")],
    ) -> dict[str, str]:
        return {
            "thought": thought,
        }

    def init_tool(self) -> Tool:
        return Tool(
            function=self._think,
            name="thinking",
            description="Use the tool to think about something. It will not obtain new information or change the database, but just append the thought to the log. Use it when complex reasoning or some cache memory is needed.",
        )


@hookimpl
def register(manager):
    manager.register(ThinkingTool())



import sys
from datetime import datetime
from functools import partial
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

_HERE = Path(__file__).parent
pwd_env = Environment(loader=FileSystemLoader(Path.cwd()), autoescape=True)
env = Environment(loader=FileSystemLoader(_HERE / "templates"), autoescape=True)

SYSTEM_PROMPT_TEMPLATE = "system_prompt.md"
CONTEXT_TEMPLATE = "context.md"


def render_template(template_file_name: str, **kwargs) -> str:
    try:
        template = pwd_env.get_template(template_file_name)
    except Exception:
        template = env.get_template(template_file_name)
    return template.render(**kwargs)


render_context = partial(render_template, CONTEXT_TEMPLATE)
render_system_prompt = partial(render_template, SYSTEM_PROMPT_TEMPLATE)


def get_system_prompt(**kwargs) -> str:
    return f"""{render_system_prompt(**kwargs)}
<Context>
{get_context()}
</Context>
"""


def get_context() -> str:
    return render_context(
        date=datetime.now().strftime("%Y-%m-%d"),
        cwd=Path.cwd().resolve().absolute().as_posix(),
        platform=sys.platform,
    )



from pydantic_ai import Agent, Tool

from lightblue_ai.models import infer_model
from lightblue_ai.settings import Settings
from lightblue_ai.tools.base import LightBlueTool, Scope
from lightblue_ai.tools.extensions import hookimpl
from lightblue_ai.tools.manager import LightBlueToolManager


class DispatchAgentTool(LightBlueTool):
    def __init__(self):
        self.settings = Settings()
        self.scopes = [Scope.exec]
        self.description = """Launch a new agent that has access to the following tools: GlobTool, GrepTool, LS, View.

When you are searching for a keyword or file and are not confident that you will find the right match on the first try, use this tool to perform the search for you. For example:

- If you are searching for a keyword like "config" or "logger", this tool is appropriate.
- If you want to read a specific file path, use the View or GlobTool tool instead to find the match more quickly.
- If you are searching for a specific class definition like "class Foo", use the GlobTool tool instead to find the match more quickly.

Usage notes:

1. Launch multiple agents concurrently whenever possible to maximize performance; to do that, use a single message with multiple tool uses.
2. When the agent is done, it will return a single message back to you. The result returned by the agent is not visible to the user. To show the user the result, you should send a text message back to the user with a concise summary of the result.
3. Each agent invocation is stateless. You will not be able to send additional messages to the agent, nor will the agent be able to communicate with you outside of its final report. Therefore, your prompt should contain a highly detailed task description for the agent to perform autonomously, and you should specify exactly what information the agent should return in its final and only message to you.
4. The agent's outputs should generally be trusted.
5. IMPORTANT: The agent cannot use Bash, Replace, Edit, so it cannot modify files. If you need to use these tools, use them directly instead of going through the agent.
"""

    async def _dispatch_agent(self, system_prompt: str, objective: str) -> str:
        tools = LightBlueToolManager().get_read_tools()

        self.agent = Agent(
            infer_model(self.settings.sub_agent_model or self.settings.default_model),
            system_prompt=system_prompt,
            tools=tools,
        )

        return await self.agent.run(objective)

    def init_tool(self) -> Tool:
        return Tool(
            name="dispatch_agent",
            function=self._dispatch_agent,
            description=self.description,
        )


@hookimpl
def register(manager):
    manager.register(DispatchAgentTool())



import enum
from abc import ABC, abstractmethod

from pydantic_ai.tools import Tool


class Scope(str, enum.Enum):
    read = "read"
    """Provide context for llm, may write to disk to store context."""

    write = "write"
    """Write, may overwrite existing files"""

    exec = "exec"
    """Exec some command, most dangerous"""

    generation = "generation"
    """AI generation context"""

    web = "web"
    """Web context"""


class LightBlueTool(ABC):
    scopes: list[Scope]

    @abstractmethod
    def init_tool(self) -> Tool:
        """Initialize the pydantic-ai tool"""

    def is_read_tool(self) -> bool:
        return Scope.read in self.scopes

    def is_write_tool(self) -> bool:
        return Scope.write in self.scopes

    def is_exec_tool(self) -> bool:
        return Scope.exec in self.scopes

    def is_generation_tool(self) -> bool:
        return Scope.generation in self.scopes

    def is_web_tool(self) -> bool:
        return Scope.web in self.scopes



import importlib
import threading
from pathlib import Path

import pluggy
from pydantic_ai.tools import Tool

import lightblue_ai.tools as tools_package
from lightblue_ai.log import logger
from lightblue_ai.tools import extensions
from lightblue_ai.tools.base import LightBlueTool
from lightblue_ai.tools.extensions import project_name as PROJECT_NAME


class Singleton(type):
    """
    metaclass for singleton, thread-safe.
    """

    _instances = {}  # noqa: RUF012
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)  # noqa: UP008
        return cls._instances[cls]


class LightBlueToolManager(metaclass=Singleton):
    def __init__(self):
        self.pm = pluggy.PluginManager(PROJECT_NAME)
        self.pm.add_hookspecs(extensions)
        self._registed_instance: list[LightBlueTool] = []

        self.load_all_local_model()
        self.pm.load_setuptools_entrypoints(PROJECT_NAME)

        self._init_tools()

    def _init_tools(self):
        for f in self.pm.hook.register(manager=self):
            try:
                f()
            except Exception as e:
                logger.exception(f"Cannot register tool {f}: {e}")

    def load_all_local_model(self):
        """
        loads all local models by automatically discovering all subdirectories in the tools directory
        """

        # Get the path of the tools directory
        tools_path = Path(tools_package.__path__[0])

        # Find all subdirectories in the tools directory
        for item in tools_path.iterdir():
            if item.is_dir() and not item.name.startswith("__"):
                # Import the module and load it
                module_name = f"lightblue_ai.tools.{item.name}"
                try:
                    module = importlib.import_module(module_name)
                    logger.debug(f"Auto-loading module: {module_name}")
                    self._load_dir(module)
                except ImportError as e:
                    logger.warning(f"Failed to import module {module_name}: {e}")

    def _load_dir(self, module):
        """
        Import all python files in a submodule.
        """
        modules = list(Path(module.__path__[0]).glob("*.py"))
        sub_packages = (p.stem for p in modules if p.is_file() and p.name != "__init__.py")
        packages = (str(module.__package__) + "." + i for i in sub_packages)
        for p in packages:
            logger.debug(f"loading {p}")
            self.pm.register(importlib.import_module(p))

    def register(self, instance: LightBlueTool):
        """
        Register a new model, if the model is already registed, skip it.
        """
        if instance in self._registed_instance:
            return
        logger.info(f"Registering tool: {instance}")
        self._registed_instance.append(instance)

    def get_sub_agent_tools(self) -> list[Tool]:
        """read&web tools"""
        return [i.init_tool() for i in self._registed_instance if i.is_read_tool() or i.is_web_tool()]

    def get_read_tools(self) -> list[Tool]:
        return [i.init_tool() for i in self._registed_instance if i.is_read_tool()]

    def get_write_tools(self) -> list[Tool]:
        return [i.init_tool() for i in self._registed_instance if i.is_write_tool()]

    def get_exec_tools(self) -> list[Tool]:
        return [i.init_tool() for i in self._registed_instance if i.is_exec_tool()]

    def get_generation_tools(self) -> list[Tool]:
        return [i.init_tool() for i in self._registed_instance if i.is_generation_tool()]

    def get_all_tools(self) -> list[Tool]:
        return [i.init_tool() for i in self._registed_instance]


if __name__ == "__main__":
    LightBlueToolManager()



import asyncio
import functools
from pathlib import Path

import typer

from lightblue_ai.agent import LightBlueAgent
from lightblue_ai.log import logger
from lightblue_ai.mcps import get_mcp_servers

app = typer.Typer()


def make_sync(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper


@app.command()
@make_sync
async def submit(
    prompt: str = typer.Argument("prompt.md", help="The prompt to send to the agent, text or file"),
    all_messages_json: Path = typer.Option(
        default="all_messages.json",
        help="The path to store the result",
    ),
):
    if Path(prompt).exists():
        with open(prompt) as f:
            prompt = f.read()

    agent = LightBlueAgent()

    result = await agent.run(prompt)
    print(result.data)

    with all_messages_json.open("wb") as f:
        f.write(result.all_messages_json())

    print(f"Usage: {result.usage()}")
    print(f"Saved all messages to {all_messages_json}")


@app.command()
def status():
    agent = LightBlueAgent()

    logger.info(f"Found {len(agent.tool_manager.get_all_tools())} tools.")
    logger.info(f"Found {len(get_mcp_servers())} MCP servers.")



from typing import Annotated, Any, Literal

from pydantic import Field
from pydantic_ai import Tool
from tavily import AsyncTavilyClient

from lightblue_ai.settings import Settings
from lightblue_ai.tools.base import LightBlueTool, Scope
from lightblue_ai.tools.extensions import hookimpl


class TavilyTool(LightBlueTool):
    def __init__(self):
        self.settings = Settings()
        self.scopes = [Scope.web]
        self.description = """Performs web searches using Tavily.
If the initial query is too broad or results are not ideal, the LLM can refine the search by progressively reducing keywords to improve accuracy.
Useful for retrieving up-to-date information, specific data, or detailed background research.
"""

    async def _search_with_tavily(
        self,
        query: Annotated[str, Field(description="The search query")],
        search_deep: Annotated[
            Literal["basic", "advanced"],
            Field(default="basic", description="The search depth"),
        ] = "basic",
        topic: Annotated[
            Literal["general", "news"],
            Field(default="general", description="The topic"),
        ] = "general",
        time_range: Annotated[
            Literal["day", "week", "month", "year", "d", "w", "m", "y"] | None,
            Field(default=None, description="The time range"),
        ] = None,
    ) -> list[dict[str, Any]]:
        client = AsyncTavilyClient(self.settings.tavily_api_key)
        results = await client.search(query, search_depth=search_deep, topic=topic, time_range=time_range)
        if not results["results"]:
            return {
                "success": False,
                "error": "No search results found.",
            }
        return results["results"]

    def init_tool(self) -> Tool:
        return Tool(
            function=self._search_with_tavily,
            name="search_with_tavily",
            description=self.description,
        )


@hookimpl
def register(manager):
    settings = Settings()
    if settings.tavily_api_key:
        manager.register(TavilyTool())



from pathlib import Path
from typing import Annotated, Any

from pdf2image import convert_from_path
from pydantic import Field
from pydantic_ai import Tool

from lightblue_ai.tools.base import LightBlueTool, Scope
from lightblue_ai.tools.extensions import hookimpl


class Pdf2ImageTool(LightBlueTool):
    """To use this tool, you need to install poppler via `brew install poppler`"""

    def __init__(self):
        self.scopes = [Scope.read]
        self.description = (
            "Converts a PDF file to a PNG image file. "
            "The file_path parameter must be an absolute path to a PDF file. "
            "The output_path parameter is optional and will default to the same directory as the input file if not provided. "
        )

    async def _pdf2images(
        self,
        file_path: Annotated[str, Field(description="Absolute path to the PDF file to convert")],
        output_path: Annotated[
            str | None,
            Field(
                description="Optional. Absolute path to the directory to save the images. "
                "If not provided, the images will be saved in the same directory as the PDF file."
            ),
        ],
    ) -> dict[str, Any]:
        try:
            path = Path(file_path).expanduser()
            if not path.exists():
                return {
                    "error": f"File not found: {file_path}",
                    "success": False,
                }

            # Check if output_path is provided
            if output_path:
                output_path = Path(output_path).expanduser()
                if not output_path.exists():
                    output_path.mkdir(parents=True, exist_ok=True)
            else:
                # If output_path is not provided, use the same directory as the input file
                output_path = path.parent

            # Convert PDF to images
            files = []
            images = convert_from_path(path, dpi=300)

            # Save each page as a PNG file
            for i, image in enumerate(images):
                image.save(f"{output_path}/output_page_{i + 1}.png", "PNG")
                files.append(f"{output_path}/output_page_{i + 1}.png")

            return {
                "success": True,
                "files": files,
                "message": f"PDF converted to {len(images)} images successfully.",
            }
        except Exception as e:
            return {
                "error": f"Failed to convert PDF to images: {e!s}",
                "success": False,
            }

    def init_tool(self) -> Tool:
        return Tool(
            function=self._pdf2images,
            name="PDF2Image",
            description=self.description,
        )


@hookimpl
def register(manager):
    manager.register(Pdf2ImageTool())



from pydantic_ai.models import infer_model as legacy_infer_model

from lightblue_ai.models.bedrock import BedrockConverseModel as PatchedBedrockConverseModel


def infer_model(model: str):
    if "bedrock:" in model:
        return PatchedBedrockConverseModel(model.lstrip("bedrock:"))
    return legacy_infer_model(model)
